import json
import os
import pymongo
import itertools

from tenant_service import TenantService
from persistance_service import PersistanceService
from job import Job

class TenantAwareJob(Job):
    
    tenant_service: TenantService

    def __init__(
        self, 
        tenant_service: TenantService, 
        persistance_service: PersistanceService
    ):
        super().__init__(persistance_service)

        self.tenant_service = tenant_service
    
    
    def run(self):
        tenants = self.tenant_service.list_tenants()
        
        for tenant in tenants:
            self._on_process_tenant(tenant)
        
        groups_of_tenants = itertools.groupby(
            sorted(tenants, key = self._get_tenant_group_key), 
            key = self._get_tenant_group_key
        )

        for group_key, group_of_tenants in groups_of_tenants:
            _group_of_tenants = list(group_of_tenants)
            self._on_process_group_of_tenants(group_key, _group_of_tenants)
        
        self._on_process_all_tenants(tenants)
        self._on_process_master_data(self.tenant_service.mongo_db_cfg['master_db'], 
                                     self.tenant_service.mongo_db_cfg['master_internal_name'])


    def _on_process_tenant(
        self, 
        tenant: dict
    ):
        pass

    
    def _get_tenant_group_key(
        self, 
        tenant: dict
    ) -> object:
        return tenant['InternalName']

    
    def _on_process_group_of_tenants(
        self, 
        tenant_group: object, 
        tenants: list
    ):
        pass


    def _on_process_all_tenants(
        self, 
        tenants: list
    ):
        pass


    def _on_process_master_data(
        self, 
        master_db: str,
        master_internal_name: str
    ):
        pass


import json
import os
import pymongo
import itertools

import requests

class RecmanService(object):
    
    base_url: str = 'https://api.recman.no/v2'

    def __init__(
        self
    ):
        super().__init__()
   
   
    def get_data(
        self, 
        key: str, 
        scope: str, 
        data_key: str,
        fields: str = None, 
        page: int = None, 
        updated: str = None,
        **extra_params: dict
    ):
        print(extra_params)
        print(extra_params.items())

        url = f"{self.base_url}/get/?key={key}&scope={scope}" \
            + (f"&fields={fields}" if fields else '') \
            + (f"&page={page}" if page else '') \
            + (f"&updated={updated}" if updated else '') \
            + '&' + '&'.join([f"{name}={value}" for name, value in extra_params.items()])
        
        response = requests.request("GET", url)

        if (response.status_code == 200):
            data = response.json()
            if data.get('success', None):
                if (type(data[data_key]) is dict):
                    return (list(data[data_key].values()), data.get('nextPage', None))
                else:
                    return (data[data_key], data.get('nextPage', None))
            elif data.get('numRows', None) == 0 and not data.get('errors', None):
                return [], None
            else:
                raise Exception(f"An error occurred during getting the data from recman. \
                    SatusCode: {response.status_code} Reason: {response.reason} Url: {url} Data: {data}")

        raise Exception(f"An error occurred during getting the data from recman. \
            SatusCode: {response.status_code} Reason: {response.reason} Url: {url}")


from pyspark.sql import *
from pyspark.sql.functions import *
from pyspark.sql.types import *

def rec_ren(
    schema, 
    rename_fn
):
    struct_fields = list()
  
    for field in schema.fields:
        if type(field.dataType) is ArrayType:
            if (type(field.dataType.elementType) is StructType):
                struct_fields.append(
                    StructField(
                        rename_fn(field.name), 
                        ArrayType(rec_ren(field.dataType.elementType, rename_fn)), 
                        field.nullable
                    )
                )
            else:
                struct_fields.append(
                    StructField(
                        rename_fn(field.name), 
                        ArrayType(field.dataType.elementType), 
                        field.nullable
                    )
                )
      
        elif type(field.dataType) is StructType:
            struct_fields.append(
                StructField(
                    rename_fn(field.name), 
                    rec_ren(field.dataType, rename_fn), 
                    field.nullable
                )
            )
      
        else:
            struct_fields.append(
                StructField(
                    rename_fn(field.name), 
                    field.dataType, 
                    field.nullable
                )
            )
  
    return StructType(struct_fields)


import json
import os
import pymongo
import itertools

class TenantService(object):
    
    mongo_db_cfg: dict

    mongo_client: pymongo.MongoClient
    mongo_master_db: pymongo.database.Database
    mongo_tenants_colection: pymongo.collection.Collection

    conn_str: str

    def __init__(
        self, 
        mongo_db_cfg: dict
    ):
        super().__init__()

        self.mongo_db_cfg = mongo_db_cfg

        self.conn_str = self.get_conn_str()

        self.mongo_client = pymongo.MongoClient(self.conn_str)
        self.mongo_master_db = self.mongo_client.get_database(self.mongo_db_cfg['master_db'])
        
        self.mongo_tenants_colection = self.mongo_master_db.get_collection('Tenants')    


    def get_conn_str(self):
        return \
            f"mongodb://" \
                + f"{self.mongo_db_cfg['user_name']}" + ':' \
                + f"{self.mongo_db_cfg['user_password']}" + '@' \
                + f"{self.mongo_db_cfg['dns_srv']}" + '/' \
                + "?authSource=admin&connectTimeoutMS=300000"
    
   
    def list_tenants(self) -> list:

        recman_key = os.environ.get("RECMAN_KEY")
        tenants = list(self.mongo_tenants_colection.find({}))
        for tenant in tenants:
            if tenant['RecmanService']['IsUsed'] == True:
                tenant['RecmanService']['ReadKey'] = recman_key

        return tenants


import json
import requests
import datetime
import pytz
import dateutil.parser

class AuthToken:
    token: str
    token_expiration_time: datetime.datetime
    auth_headers: str

    def __init__(self, token: str, token_expiration_time: datetime.datetime):
        self.token = token
        self.token_expiration_time = token_expiration_time
        self.auth_headers = {'Content-Type': "application/json",'Authorization': token}


class BackendService:

    base_url: str
    password: str
    auth_token_cache: dict

    def __init__(self, base_url: str, password: str):
        self.auth_token_cache = dict()
        self.base_url = base_url
        self.password = password

    def __get_auth_token__(self, tenant_id: str) -> AuthToken:
        url = self.base_url + "/api/Auth/AuthUser/"
        payload = json.dumps({ "name": tenant_id, "password": self.password})
        headers = { 'Content-Type': "application/json" }
        response = requests.request("POST", url, data = payload, headers = headers)
        
        if (not response.ok):
            raise Exception(f"Can not get an access token. \
                SatusCode: {response.status_code} Reason: {response.reason}")
        
        token = response.json()['_Token']
        token_expiration_time = dateutil.parser.parse(response.json()['expirationDate'])

        return AuthToken(token, token_expiration_time)
        
    def __get_auth_token_from_cache__(self, tenant_id: str):
        auth_token = self.auth_token_cache.get(tenant_id)
        if (auth_token is None):
            auth_token = self.__get_auth_token__(tenant_id)
        
        current_utc_time = datetime.datetime.now(tz=pytz.utc)
        corrected_expiration_time = auth_token.token_expiration_time - datetime.timedelta(minutes=15)
        if (current_utc_time > corrected_expiration_time):
            auth_token = self.__get_auth_token__(tenant_id)

        self.auth_token_cache[tenant_id] = auth_token

        return auth_token
    


import json
import os
import pymongo
import itertools

from pyspark.sql import *
from pyspark.sql.window import *
from pyspark.sql.functions import *
from pyspark.sql.types import *
from pyspark.dbutils import DBUtils

from tenant_service import TenantService
from persistance_service import PersistanceService
from tenant_aware_job import TenantAwareJob
from spark import rec_ren

class TenantAwareSparkJob(TenantAwareJob):
    spark_config: dict
    spark: SparkSession

    def __init__(
        self,
        tenant_service: TenantService, 
        persistance_service: PersistanceService,
        spark_cfg: dict,
        job_name: str
    ):
        super().__init__(tenant_service, persistance_service)

        self.spark_cfg = spark_cfg

        SparkSession.builder.config("spark.app.id", job_name)
        SparkSession.builder.config("spark.databricks.service.address", self.spark_cfg["address"])
        SparkSession.builder.config("spark.databricks.service.token", self.spark_cfg["token"])
        SparkSession.builder.config("spark.databricks.service.clusterId", self.spark_cfg["clusterId"])
        SparkSession.builder.config("spark.databricks.service.orgId", self.spark_cfg["orgId"])
        SparkSession.builder.config("spark.databricks.service.port", self.spark_cfg["port"])
        self.spark = SparkSession.builder.getOrCreate()


    def _enforce_naming_notation(
        self,
        df: DataFrame
    ) -> DataFrame:
        return self.spark \
            .createDataFrame(
                df.rdd,
                rec_ren(df.schema, lambda x: x[:1].upper() + x[1:])
            )
    
    
    def _save_data(
        self,
        df: DataFrame,
        tenant: dict,
        collection: str,
        write_mode: str = "append"
    ):
        df \
            .repartition(3) \
            .write \
            .format("com.mongodb.spark.sql.DefaultSource") \
            .mode(write_mode) \
            .option("uri", self.tenant_service.get_conn_str()) \
            .option("database", tenant['Database']) \
            .option("collection", collection) \
            .option("replaceDocument", "true") \
            .save()
        

    def _collapse_to_single_file(
        self,
        old_file_path_prefix: str, 
        new_file_path_prefix: str
    ):
        dbutils = DBUtils(self.spark.sparkContext)
        
        old_file_paths = dbutils.fs.ls(f"{old_file_path_prefix}/")

        partition_path = filter(
            lambda f: f.name.startswith('part-'),
            old_file_paths
        )
        
        partition_path = next(partition_path)

        if not partition_path:
            raise Exception(f"The file path {old_file_path_prefix} does not contain any partitions")
        else:
            partition_path = partition_path.path
        
        dbutils.fs.cp(
            partition_path, 
            new_file_path_prefix
        )

        dbutils.fs.rm(
            old_file_path_prefix, 
            recurse = True
        )
    


import json
import os
import pymongo
import itertools
from azure.storage.blob import BlockBlobService

class PersistanceService(object):
    
    blob_storage_cfg: dict
    blob_service: BlockBlobService

    def __init__(
        self, 
        blob_storage_cfg: dict
    ):
        super().__init__()

        self.blob_storage_cfg = blob_storage_cfg
        
        self.blob_service = BlockBlobService(
            blob_storage_cfg['account_name'], 
            blob_storage_cfg['account_key']
        )   

    
    def save_data(
        self, 
        name: str, 
        data: str
    ):
        self.blob_service.create_blob_from_text(
            self.blob_storage_cfg['container_name'], 
            name, 
            data
        )


    def save_data_from_file(
        self, 
        name: str,
        file_path: str
    ):
        self.blob_service.create_blob_from_path(
            self.blob_storage_cfg['container_name'], 
            name,
            file_path
        )
        

    def get_data(
        self, 
        name: str
    ) -> str:

        blob = self.blob_service.get_blob_to_text(
            self.blob_storage_cfg['container_name'], 
            name
        )

        return blob.content

    
    def get_data_if_exists(
        self, 
        name: str
    ) -> str:

        container_name = self.blob_storage_cfg['container_name']

        if self.blob_service.exists(container_name, name):
            return self.get_data(name)
        else:
            return None


    def list_items(
        self, 
        prefix: str
    ) -> str:

        blobs = self.blob_service.list_blobs(
            self.blob_storage_cfg['container_name'],
            prefix = prefix
        )

        return [b.name for b in blobs]


import json
import os
import pymongo
import itertools

from tenant_service import TenantService
from persistance_service import PersistanceService

class Job(object):
    
    persistance_service: PersistanceService

    def __init__(
        self, 
        persistance_service: PersistanceService
    ):
        super().__init__()

        self.persistance_service = persistance_service
   

    def run(self):
        pass


    def _get_save_point_file_name(
        self, 
        recman_key: str, 
        scope_name: str
    ) -> str:
        return 'savepoint.json'


    def _get_save_point(
        self, 
        recman_key: str, 
        scope_name: str, 
        default_value: dict
    ) -> dict:
        file_name = self._get_save_point_file_name(recman_key, scope_name)

        save_point = self.persistance_service.get_data_if_exists(file_name)
        save_point = save_point and json.loads(save_point) or default_value

        return save_point
    

    def _persist_save_point(
        self, 
        recman_key: str, 
        scope_name: str, 
        value: dict
    ):
        file_name = self._get_save_point_file_name(recman_key, scope_name)

        self.persistance_service.save_data(file_name, json.dumps(value))



import pyodbc

def exec_sql_srv_proc(sql_srv: str, db_name: str, usr_name: str, usr_pass: str, proc_name: str):
    conn = pyodbc.connect(
        'DRIVER={ODBC Driver 17 for SQL Server};'
        f"SERVER={sql_srv};"
        f"DATABASE={db_name};UID={usr_name};"
        f"PWD={usr_pass}"
    )

    conn.execute(f"EXEC {proc_name}")
    conn.commit()

    conn.close()


import json

def list_to_json_line(l: list) -> str:
    return '\n'.join([json.dumps(item) for item in l])


def json_line_to_list(jsl: str) -> str:
    return [json.loads(line) for line in jsl.split('\n') if line]


import json
import os
import pymongo
import itertools

from job import Job
from tenant_service import TenantService
from persistance_service import PersistanceService

class Pipeline(Job):
    
    pipeline: list

    def __init__(
        self,
        persistance_service: PersistanceService ,
        *args
    ):
        super().__init__(persistance_service)

        self.pipeline = args
   
   
    def run(self):
        for job in self.pipeline:
            job.run()
        
        pass


import json
import os
import requests
from time import sleep
from backend_service import BackendService

class BackendCandidatesService(BackendService):

    def send(self, tenant_id: str, data: list) -> dict:
        auth_token = self.__get_auth_token_from_cache__(tenant_id)

        url = f"{self.base_url}/api/Candidates/UpsertCandidates"

        for i in range(3):
            response = requests.request("POST", url, json = data, headers = auth_token.auth_headers)
            
            print(response.status_code, response.content)
            
            if response.status_code != 200:
                sleep(i * 60)
                continue
            break

        if (not response.ok):
            raise Exception(f"An error occurred during posting candidates to the backend. \
                SatusCode: {response.status_code} Reason: {response.reason}")


import json
import os

class ConfigReader:
    
    working_dir: str
    app_name: str

    config: dict

    def __init__(
        self, 
        working_dir: str, 
        app_name: str
    ):
        super().__init__()

        self.working_dir = working_dir
        self.app_name = app_name

        self.read()

    
    def read(self):
        env_name = os.environ.get('ENV_NAME', "dev")
        config_file_path = f"{self.working_dir}/{self.app_name}.{env_name}.cfg.json"
        
        with open(config_file_path) as json_file:
            self.config = json.load(json_file)
        
        return


import json
import os
import itertools

from pyspark.sql import *
from pyspark.sql.window import *
from pyspark.sql.functions import *
from pyspark.sql.types import *
from pyspark.dbutils import DBUtils

from persistance_service import PersistanceService
from job import Job

class SparkJob(Job):

    spark_config: dict
    spark: SparkSession
    persistance_service: PersistanceService

    def __init__(
        self,
        persistance_service: PersistanceService,
        spark_cfg: dict,
        job_name: str
    ):
        super().__init__(persistance_service)

        self.spark_cfg = spark_cfg

        SparkSession.builder.config("spark.app.id", job_name)
        SparkSession.builder.config("spark.databricks.service.address", self.spark_cfg["address"])
        SparkSession.builder.config("spark.databricks.service.token", self.spark_cfg["token"])
        SparkSession.builder.config("spark.databricks.service.clusterId", self.spark_cfg["clusterId"])
        SparkSession.builder.config("spark.databricks.service.orgId", self.spark_cfg["orgId"])
        SparkSession.builder.config("spark.databricks.service.port", self.spark_cfg["port"])
        self.spark = SparkSession.builder.getOrCreate()

        
    def _save_table_to_csv(self, table_name: str, file_path: str):
        self.spark \
            .read \
            .table(table_name) \
            .repartition(1) \
            .write \
            .mode("overwrite") \
            .option("header", "true") \
            .csv(file_path)
    
    


from setuptools import setup, find_packages

with open('README.md') as readme_file:
    README = readme_file.read()

setup_args = dict(
    name='globus.etl.utils',
    version='0.0.2',
    description='ETL tools for Globus Staffing',
    #package_dir={'': 'src'},
    #long_description_content_type="text/markdown",
    #long_description=README + '\n\n' + HISTORY,
    license='MIT',
    packages=find_packages(),
    author='globus.ai',
    author_email='zhas@globus.ai',
    keywords=['Staffing', 'ETL'],
    url='https://bitbucket.org/dedicare/globus_etl_utils/src/develop/',
    download_url='https://pypi.org/project/globusetlutils/'
)

install_requires = [
    'pytz>=2019.3',
    'pyodbc>=4.0.30',
    'pymongo>=3.8.0',
    'requests>=2.23.0',
    'azure-storage-blob>=0.37.0',
    'pyspark>=3.0.0',
    'python_dateutil>=2.8.1',
]

if __name__ == '__main__':
    setup(**setup_args, install_requires=install_requires)